##################################################################################
#                                                                                #
# This script infers causal relationships between genetic mutation               #
# events from co-occurrence tables.                                              #
#                                                                                #
# Authors: Paola Lecca, Francesca Demichelis, Nicola Casiraghi,                  #
#          CIBIO, University of Trento, Italy                                    #
# Contact: P. Lecca: paola.lecca@unitn.it                                        #
#          F. Demichelis: f.demichelis@unitn.it                                  #
#          N. Casiraghi: nicola.casiraghi@unitn.it                               #
#                                                                                #
# Usage: Rscript TO-DAG.R "./input.txt" 0.03 outputdir                           #
# Requirements: utils, igraph                                                    #
#                                                                                #
#                                                                                #
##################################################################################  

# loading required packages
if (!require(utils, quietly = T, warn.conflicts = F)) {
  stop("utils R package is not installed.\n")
}
if (!require(igraph, quietly = T, warn.conflicts = F)) {
  stop("igraph R package is not installed.\n")
}

# enable command line options
Args <- commandArgs(TRUE)
if (length(Args) > 2) {
  mainDir <- normalizePath(Args[3], mustWork = F)
  if (!file.exists(mainDir)) {
    stop(paste("Output folder", mainDir, "does not exist.\n"))
  }
} else {
  mainDir <- getwd()  
}

# Input data
events_data <- (as.matrix(read.table(Args[1], header=TRUE)))[, -1]

class(events_data) <- "numeric"

n_events <- dim(events_data)[2]
n_samples <- dim(events_data)[1]

short_gene_names <- colnames(events_data)

# create directory where to store the output
subDir <- c("plots", "networks", "degree")

for (i in 1:length(subDir))
{
  if (file.exists(file.path(mainDir, subDir[i]))) {
    if (file.info(file.path(mainDir, subDir[i]))$isdir) {
      cat(subDir[i], "exists in", mainDir, "and is a directory.\n")      
    } else {
      stop( paste(subDir[i], "exists in", mainDir, "but is a file.\n"))
    }
  } else {
    cat(subDir[i], "does not exist in", mainDir, "- creating.\n")
    dir.create(file.path(mainDir, subDir[i]))
  }
  if (file.exists(file.path(mainDir, subDir[i]))) {
    setwd(file.path(mainDir, subDir[i]))
  } else {
    cat(subDir[i], "does not exist.\n")
  }
}

# Probability of an event
probability_event <- function(data, event)
{
  prob <- ( sum(data[, event]) )/ (dim(data)[1])
  prob
}

probability_array <- function(data)
{
  prob_e <- array(0, dim(data)[2])
  for (j in 1:dim(data)[2])
  {
    prob_e[j] <- probability-evet(data, j)
  }
  prob_e
}

# Conditional probability
intersect_prob <- function(event_1, event_2, data)
{
  count <- 0
  for (i in 1:dim(data)[1])
  {
    if ( (data[i,event_1]==1) && (data[i,event_2]==1) )
    {
      count <- count + 1
    }
  }
  inters_prob <- count/((dim(data)[1]))^2
  inters_prob
}

intersect_prob_multiple <- function(event.v, data)
{
  m1 <- data[, event.v]
  m1.rsums <- rowSums(m1)
  pos.full <- which(m1.rsums==length(event.v))
  res1 <- m1[pos.full,]
  intersect.p <- dim(res1)[1]/dim(data)[1]
  intersect.p
}

cond_prob <- function(event_1, event_2, data)
{
  cp <- (intersect_prob(event_1, event_2, data))/(probability_event(data, event_2))
  cp
}

#  matrix of the combination of all events

all_couples_events <- t(combn(1:n_events,2))

cpa_i_j <- array(0, dim(all_couples_events)[1])
cpa_j_i <- array(0, dim(all_couples_events)[1])
for (i in 1:dim(all_couples_events)[1])
{
  cpa_i_j[i]  <- (cond_prob(all_couples_events[i,1], all_couples_events[i, 2], events_data))/probability_event(events_data, all_couples_events[i, 2] )
  cpa_j_i[i]  <- (cond_prob(all_couples_events[i,2], all_couples_events[i, 1], events_data))/probability_event(events_data, all_couples_events[i, 1] )
  
}

cpa_mat <- cbind(all_couples_events[,1], all_couples_events[,2], cpa_i_j,
                 all_couples_events[,2], all_couples_events[,1], cpa_j_i)

# Establish directions

inferred_net <- matrix(0, dim(cpa_mat)[1], 3)

for (i in 1: dim(cpa_mat)[1])
{
  if(cpa_mat[i,3] > cpa_mat[i,6])
  {
    inferred_net[i, 1] <- cpa_mat[i, 1]
    inferred_net[i, 2] <- cpa_mat[i, 2]
    inferred_net[i, 3] <- cpa_mat[i, 3]
  }
  if(cpa_mat[i,3] < cpa_mat[i,6])
  {
    inferred_net[i, 1] <- cpa_mat[i, 2]
    inferred_net[i, 2] <- cpa_mat[i, 1]
    inferred_net[i, 3] <- cpa_mat[i, 6]
  }
  if(cpa_mat[i,3] == cpa_mat[i,6])
  {
    inferred_net[i, 1] <- cpa_mat[i, 1]
    inferred_net[i, 2] <- cpa_mat[i, 2]
    inferred_net[i, 3] <- 0
  }
}

inferred_net.orig <- inferred_net
inferred_net <- subset(inferred_net, inferred_net[,3] >= Args[2])

if(dim(inferred_net)[1] ==0)
{
  stop(paste("\n\n\nNo edges have weight greater or equal than", Args[2], 
             "\nThe maximum edge weight is", max(inferred_net.orig[,3])))
}

pdf(file.path(mainDir, "plots/edge_weight.pdf"))
plot(density(inferred_net[,3]), main="Edge weight distribution",
     xlab="Edge weight", cex.axis=1.4, cex.lab=1.4, cex.main=1.4, lwd=2)
invisible(dev.off())
          
if (dim(inferred_net)[1]!=0)
{
  inferred_net_names <- matrix("", dim(inferred_net)[1], 2)
  for (i in 1:dim(inferred_net)[1])
  {
    inferred_net_names[i, 1] <- short_gene_names[inferred_net[i,1]]
    inferred_net_names[i,2] <- short_gene_names[inferred_net[i,2]]
  }
  
  actors <- data.frame(name=short_gene_names)
  
  relations <- data.frame(from=inferred_net_names[, 1],
                          to=inferred_net_names[, 2])
  
  g <- graph.data.frame(relations, directed=TRUE, vertices=actors)
  V(g)$label <- V(g)$name
  
  E(g)$weight <- inferred_net[,3]
  
  # Waiting times of mutation
  den <- log( -runif(dim(inferred_net_names)[1]) + 1 )
  waiting_times <- (den/-inferred_net[,3])
  
  edges_prob_times <- cbind(inferred_net_names, inferred_net[,3], waiting_times)
  
  edges_prob_times <- subset(edges_prob_times, as.numeric(edges_prob_times[,4]) > mean(as.numeric(edges_prob_times[,4])))
  
  pdf(file = file.path(mainDir, "plots/waiting_times.pdf"))
  hist(waiting_times, main="Histogram of waiting times", xlab="waiting time", col="lightgray")
  invisible(dev.off())

  # Assign colors w.r.t to waiting times
  
  wt <- waiting_times
  q.wt <- quantile(wt)
  
  q1 <- q.wt[1]
  q2 <- q.wt[2]
  q3 <- q.wt[3]
  q4 <- q.wt[4]
  
  colrs <- array("",length(wt)) 
  for (i in 1:length(wt))
  {
    # fast
    if ((wt[i] >= 0) && (wt[i] < q2))
    {
      colrs[i] <- "green"
    }
    # mean
    if( (wt[i] >=q2) &&   (wt[i] < q4) ) 
    {
      colrs[i] <- "orange"
    }
    # medium
    if( wt[i] >= q4)
    {
      colrs[i] <- "red"
    }
  }

  E(g)$width <- 2
  E(g)$color <- colrs
  
  all.degree <- cbind( V(g)$name, degree(g, mode="all"))
  in.degree <- cbind(V(g)$name, degree(g, mode="in"))
  out.degree <- cbind(V(g)$name, degree(g, mode="out"))
  
  all.degree <- all.degree[order(as.numeric(all.degree[,2]), decreasing = T),]
  in.degree <- in.degree[order(as.numeric(in.degree[,2]), decreasing = T),]
  out.degree <- out.degree[order(as.numeric(out.degree[,2]), decreasing = T),]
  
  pdf( file.path(mainDir, "plots/degree.pdf"))
  par(mfrow=c(1,3))
  hist(as.numeric(all.degree[,2]), main="Total degree histogram", xlab="Total degree", cex.axis=1, cex.lab=1.2, col="lightgray")
  hist(as.numeric(in.degree[,2]), main="In-degree histogram", nclass=3, xlab="In-degree",cex.axis=1, cex.lab=1.2, col="lightgray")
  hist(as.numeric(out.degree[,2]), main="Out-degree histogram", xlab="out-degree",cex.axis=1, xlim=c(0,270),cex.lab=1.2, col="lightgray")
  invisible(dev.off())
  
  all.d <- summary(as.numeric(all.degree[,2]))
  in.d <- summary(as.numeric(in.degree[,2]))
  out.d <- summary(as.numeric(out.degree[,2]))
  
  degree.hist <- data.frame(rbind(all.d, in.d, out.d))
  rownames(degree.hist) <- c("Total degree", "In-degree", "Out-degree")
  
  write.table(degree.hist, file = file.path(mainDir, "degree/degree_hist_summary.txt"), quote = F)
  
  write(t(all.degree), file = file.path(mainDir, subDir[3], "network_all_degree.txt"), ncol=2, sep="\t")
  write(t(in.degree), file = file.path(mainDir, subDir[3], "network_in_degree.txt"), ncol=2, sep="\t")
  write(t(out.degree), file = file.path(mainDir, subDir[3], "network_tot_degree.txt"), ncol=2, sep="\t")
  
  write.graph(g, file = file.path(mainDir, subDir[2],"inferred_graph.graphml"), format="graphml")
  
  sumTable.net <- cbind(get.edgelist(g), E(g)$color, E(g)$weight)
  write(t(sumTable.net), file = file.path(mainDir, subDir[2], "inferred_graph.txt"), ncol=4, sep="\t")
  
  edges.red <- E(g)[which(E(g)$color=="red")]
  edges.green <- E(g)[which(E(g)$color=="green")]
  edges.orange <- E(g)[which(E(g)$color=="orange")]
  
  red.sg <- subgraph.edges(g, eids=as.vector(edges.red))
  green.sg <- subgraph.edges(g, eids=as.vector(edges.green))
  orange.sg <- subgraph.edges(g, eids=as.vector(edges.orange))
  
  write.graph(red.sg, file = file.path(mainDir, subDir[2], "red_edges_graph.graphml"), format="graphml")
  write.graph(green.sg, file= file.path(mainDir, subDir[2], "green_edges_graph.graphml"), format="graphml")
  write.graph(orange.sg, file= file.path(mainDir, subDir[2], "orange_edges_graph.graphml"), format="graphml")
  
  red.sif <- cbind(get.edgelist(red.sg), E(red.sg)$weight)  
  green.sif <- cbind(get.edgelist(green.sg), E(green.sg)$weight)  
  orange.sif <- cbind(get.edgelist(orange.sg), E(orange.sg)$weight)  
  
  write(t(red.sif), file = file.path(mainDir, subDir[2], "red_edges_SIF.txt"), sep="\t", ncol=3)
  write(t(green.sif), file = file.path(mainDir, subDir[2], "green_edges_SIF.txt"), sep="\t", ncol=3)
  write(t(orange.sif), file = file.path(mainDir, subDir[2], "orange_edges_SIF.txt"), sep="\t", ncol=3)
  
  # degree of the nodes in each subgraph
  red.degree.all <- cbind( V(red.sg)$name, degree(red.sg, mode="all") )
  green.degree.all <- cbind( V(green.sg)$name, degree(green.sg, mode="all") )
  orange.degree.all <- cbind( V(orange.sg)$name, degree(orange.sg, mode="all") )
  
  red.degree.in <- cbind( V(red.sg)$name, degree(red.sg, mode="in"))
  green.degree.in <- cbind( V(green.sg)$name, degree(green.sg, mode="in"))
  orange.degree.in <- cbind( V(orange.sg)$name, degree(orange.sg, mode="in"))
  
  red.degree.out <- cbind( V(red.sg)$name,degree(red.sg, mode="out") )
  green.degree.out <- cbind( V(green.sg)$name,degree(green.sg, mode="out"))
  orange.degree.out <- cbind( V(orange.sg)$name,degree(orange.sg, mode="out"))
  
  # degree of nodes in each subgraph increasingly sorted
  red.degree.all <- red.degree.all[order(as.numeric(red.degree.all[,2]), decreasing = T),]
  green.degree.all <- green.degree.all[order(as.numeric(green.degree.all[,2]), decreasing = T),]
  orange.degree.all <- orange.degree.all[order(as.numeric(orange.degree.all[,2]), decreasing = T),]
  
  red.degree.in <- red.degree.in[order(as.numeric(red.degree.in[,2]), decreasing = T),]
  green.degree.in <- green.degree.in[order(as.numeric(green.degree.in[,2]), decreasing = T),]
  orange.degree.in <- orange.degree.in[order(as.numeric(orange.degree.in[,2]), decreasing = T),]
  
  red.degree.out <- red.degree.out[order(as.numeric(red.degree.out[,2]), decreasing = T),]
  green.degree.out <- green.degree.out[order(as.numeric(green.degree.out[,2]), decreasing = T),]
  orange.degree.out <- orange.degree.out[order(as.numeric(orange.degree.out[,2]), decreasing = T),]
  
  write(t(red.degree.all), file = file.path(mainDir, subDir[3], "all_degree_sequence_red.txt"), ncol=2, sep="\t")
  write(t(green.degree.all), file = file.path(mainDir, subDir[3], "all_degree_sequence_green.txt"), ncol=2, sep="\t")
  write(t(orange.degree.all), file = file.path(mainDir, subDir[3], "all_degree_sequence_orange.txt"), ncol=2, sep="\t")
  
  write(t(red.degree.in), file = file.path(mainDir, subDir[3], "in_degree_sequence_red.txt"), ncol=2, sep="\t")
  write(t(green.degree.in), file = file.path(mainDir, subDir[3], "in_degree_sequence_green.txt"), ncol=2, sep="\t")
  write(t(orange.degree.in), file = file.path(mainDir, subDir[3], "in_degree_sequence_orange.txt"), ncol=2, sep="\t")
  
  write(t(red.degree.out), file = file.path(mainDir, subDir[3], "out_degree_sequence_red.txt"), ncol=2, sep="\t")
  write(t(green.degree.out), file = file.path(mainDir, subDir[3], "out_degree_sequence_green.txt"), ncol=2, sep="\t")
  write(t(orange.degree.out), file = file.path(mainDir, subDir[3], "out_degree_sequence_orange.txt"), ncol=2, sep="\t")
}                    

# Re-calculate the probability of mutation occurrence (with no Markov assumption)

# dominator event
freq.v <- (colSums (events_data, na.rm = FALSE, dims = 1))/dim(events_data)[1] 
max.freq <- max( (colSums (events_data, na.rm = FALSE, dims = 1))/dim(events_data)[1] )
pos.max <- which(freq.v == max.freq)

all.paths <- all_simple_paths(g, from=V(g)$label[pos.max], to = V(g), mode="out")

p.intersect <- list()
length(p.intersect) <- length(all.paths) 

for (i in 1:length(all.paths))
{
  p.intersect[[i]][1] <- 0
  for (j in 2:length(all.paths[[i]]))
  {
    p.intersect[[i]][j] <- intersect_prob_multiple( V(g)$label[all.paths[[i]][1]:all.paths[[i]][j]], events_data)
  }
}

save(all.paths, p.intersect, file = file.path(mainDir, subDir[2], "path_prob.RData"))



